"""
This is the entry point for the command-line interface (CLI) application.

It can be used as a handy facility for running the task from a command line.

.. note::

    To learn more about Click visit the
    `project website <http://click.pocoo.org/5/>`_.  There is also a very
    helpful `tutorial video <https://www.youtube.com/watch?v=kNke39OZ2k0>`_.

    To learn more about running Luigi, visit the Luigi project's
    `Read-The-Docs <http://luigi.readthedocs.io/en/stable/>`_ page.

.. currentmodule:: databutton.cli
.. moduleauthor:: Databutton <support@databutton.com>
"""
# !/usr/bin/env python
# -*- coding: utf-8 -*-
from uuid import uuid4
import requests
import tarfile
import logging
import click
from yaspin import yaspin

from databutton.utils import create_databutton_config, get_databutton_config, get_databutton_login_info, upload_archive
from databutton.utils.deploy import listen_to_build
from .__init__ import __version__
import os
import webbrowser
from databutton.utils.build import create_archive

LOGGING_LEVELS = {
    0: logging.NOTSET,
    1: logging.ERROR,
    2: logging.WARN,
    3: logging.INFO,
    4: logging.DEBUG,
}  #: a mapping of `verbose` option counts to logging levels


class Info(object):
    """An information object to pass data between CLI functions."""

    def __init__(self):  # Note: This object must have an empty constructor.
        """Create a new instance."""
        self.verbose: int = 0


# pass_info is a decorator for functions that pass 'Info' objects.
#: pylint: disable=invalid-name
pass_info = click.make_pass_decorator(Info, ensure=True)


# Change the options to below to suit the actual options for your task (or
# tasks).
@click.group()
@click.option("--verbose", "-v", count=True, help="Enable verbose output.")
@pass_info
def cli(info: Info, verbose: int):
    """Run databutton."""
    # Use the verbosity count to determine the logging level...
    if verbose > 0:
        logging.basicConfig(
            level=LOGGING_LEVELS[verbose]
            if verbose in LOGGING_LEVELS
            else logging.DEBUG
        )
        click.echo(
            click.style(
                f"Verbose logging is enabled. "
                f"(LEVEL={logging.getLogger().getEffectiveLevel()})",
                fg="yellow",
            )
        )
    info.verbose = verbose


@cli.command()
def version():
    """Get the library version."""
    click.echo(click.style(f"{__version__}", bold=True))


@cli.command()
@click.option('--template', default='sales-forecasting', type=click.Choice(['sales-forecasting']), prompt='Select a template')
@click.option('--name', default='my-databutton-app', prompt='Name of the app')
def create(template, name):
    """Create an application with a template"""
    click.echo(click.style(f"You selected {template}"))

    # Fetch repository
    os.mkdir(name)
    click.echo("Download template...")
    res = requests.get(f"https://storage.googleapis.com/databutton-app-templates/{template}.tar.gz", stream=True)
    if res.ok:
        tmpname = f"/tmp/databutton-{template}.tar.gz"
        with open(tmpname, 'wb') as f:
            f.write(res.raw.read())
        tar_obj = tarfile.open(tmpname)
        tar_obj.extractall(name)
        os.remove(tmpname)
        click.echo()
        click.echo(click.style(f"Successfully created {name}", fg='green'))
        click.echo()
        click.echo(click.style(f"cd into the folder (cd {name}) and start hacking!"))
        click.echo(click.style("Use your favorite environment setup to install requirements with"))
        click.echo()
        click.echo(click.style("pip install -r requirements.txt", fg="green"))
        click.echo()
        click.echo(click.style("write `databutton dev` and make the data app yours!", fg="green"))
        click.echo()
        click.echo()

    else:
        click.echo(click.style(f"Error, could not get template {template}"))
        quit(1)


@cli.command()
@click.option('--reload', default=True, is_flag=True, help='Hot reload')
def dev(reload):
    """Run the development setup of databutton"""
    import uvicorn
    import sys
    from uvicorn.supervisors import ChangeReload
    dir_path = os.path.dirname(os.path.realpath(__file__))
    app_dir = os.path.join(dir_path, 'server')
    sys.path.insert(0, app_dir)
    config = uvicorn.Config("server:app", reload=reload, reload_excludes=".databutton/**/*.py")
    server = uvicorn.Server(config=config)
    server.force_exit = True
    sock = config.bind_socket()
    supervisor = ChangeReload(config, target=server.run, sockets=[sock])
    import webbrowser
    webbrowser.open('http://localhost:8000')
    supervisor.run()


@cli.command()
def login():
    """Login to databutton"""
    import uvicorn
    dir_path = os.path.dirname(os.path.realpath(__file__))
    app_dir = os.path.join(dir_path, 'auth')
    click.echo(click.style("Opening browser to authenticate"))
    click.echo(click.style("Please finish the login in the browser at http://localhost:8000"))
    click.echo()
    click.echo(click.style("Exit this process with ctrl+c"))
    click.echo()
    webbrowser.open_new_tab('http://localhost:8000')
    uvicorn.run("server:app", app_dir=app_dir, reload=False, log_level='debug')


@cli.command()
@click.pass_context
def deploy(ctx: click.Context):
    """Deploy to databutton"""

    def get_auth():
        try:
            login_info = get_databutton_login_info()
            if login_info is None:
                raise Exception()
            return login_info
        except Exception:
            click.echo(click.style("No login info found, logging in first..."))
            click.echo()
            ctx.invoke(login)
            click.echo()
            return get_auth()

    def get_config():
        try:
            config = get_databutton_config()
            return config
        except Exception:
            click.echo(click.style("No databutton.json config found, creating a new one..."))
            ctx.invoke(create_databutton_project)
            return get_config()

    get_auth()  # Make sure the user has a valid auth thingy
    config = get_config()
    deployment_id = str(uuid4())

    click.echo()
    click.echo(click.style(f"=== Deploying to {config.name}", fg='green'))
    ctx.invoke(build)
    click.echo(click.style('i packaging components...', fg='cyan'))
    create_archive(deployment_id)
    click.echo(click.style("i done packaging components", fg='green'))
    click.echo(click.style("i uploading components", fg='cyan'))
    click.echo(click.style("i finished uploaded components", fg='green'))
    upload_archive(config, deployment_id)
    click.echo(click.style("i cleaning up", fg='cyan'))
    os.system(f"rm {deployment_id}")

    click.echo(click.style("i waiting for deployment to be ready, this can take some time..", fg='cyan'))
    with yaspin(text=click.style("i deploying...", fg='cyan'), color="cyan"):
        listen_to_build(deployment_id)

    click.echo(click.style("✅ Done!", fg='green'))
    click.echo()
    click.echo(click.style(f"You can now go to https://p{config.uid}.dbtn.app"))
    click.echo()


@cli.command()
def build():
    click.echo(click.style("i building project", fg='cyan'))
    from databutton.utils.build import generate_components
    click.echo(click.style("i generating components", fg='cyan'))
    generate_components()
    click.echo(click.style("i finished building project in .databutton", fg='green'))


@cli.command()
def serve():
    click.echo(click.style("=== Serving"))
    import uvicorn
    dir_path = os.path.dirname(os.path.realpath(__file__))
    app_dir = os.path.join(dir_path, 'server')
    port = os.environ['PORT'] if 'PORT' in os.environ else 8000
    uvicorn.run('prod:app', app_dir=app_dir, reload=False, port=int(port), host='0.0.0.0')


@cli.command()
@click.option('--name', default=os.path.basename(os.getcwd()), prompt=True)
def create_databutton_project(name: str):
    config = create_databutton_config(name)
    click.echo(f"Created project with name {name}, see databutton.json")
    return config
