import asyncio
import importlib
import json
import os

from databutton.decorators.dataframe import DataframeArtifact
from databutton.server.processes import forward, get_proxy, reverse
from fastapi import FastAPI, Response, WebSocket, status
from fastapi.responses import RedirectResponse
from fastapi.staticfiles import StaticFiles
from websockets import connect

from databutton.utils.config import MAX_WEBSOCKET_MESSAGE_SIZE_IN_MB

app = FastAPI()

with open('.databutton/artifacts.json', 'r') as f:
    components = json.load(f)

# Register dataframe routes
df: DataframeArtifact
for df in components['dataframes']:
    name = df['modulename']
    file = df['importfile']
    spec = importlib.util.spec_from_file_location(name, file)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)

    app.include_router(mod.router)

# Set up proxies for streamlit apps
# They are ran as containers somewhere

for st in components['streamlit_apps']:
    @app.get(st['route'] + "{rest:path}")
    async def get_app(rest: str):
        if not rest:
            return RedirectResponse(f"{st['route']}/")
        return await get_proxy(st['route'], rest, port=st['port'])

    @app.websocket(st['route'] + "/stream")
    async def handle_proxied_websocket(ws_client: WebSocket):
        await ws_client.accept()
        port = st['port']
        if port is None:
            return Response(status_code=status.HTTP_404_NOT_FOUND)
        max_size = MAX_WEBSOCKET_MESSAGE_SIZE_IN_MB * int(1e6)
        async with connect(f"ws://localhost:{port}/stream", max_size=max_size) as ws_server:
            fwd_task = asyncio.create_task(forward(ws_client, ws_server))
            rev_task = asyncio.create_task(reverse(ws_client, ws_server))
            await asyncio.gather(fwd_task, rev_task)


# @app.get('/')
# async def index():
#     return {
#         'status': 'SUCCESS',
#         'code': 200
#     }

@app.get('/')
async def index():
    return RedirectResponse(url='/index.html')

dir_path = os.path.dirname(os.path.realpath(__file__))
app_dir = os.path.join(dir_path, '../out')
app.mount('/static', StaticFiles(directory='.databutton'), name='.databutton')
app.mount('/', StaticFiles(directory=app_dir), name='app')


@app.get('/healthz')
async def healthz():
    return Response(status_code=status.HTTP_200_OK)
