from dataclasses import dataclass
from importlib.resources import path
import os
import json
from pathlib import Path
from typing import Union
import uuid
import requests

try:
    from urllib.parse import quote
except Exception:
    from urllib import quote


@dataclass
class LoginData:
    refreshToken: str
    uid: str


@dataclass
class ProjectConfig:
    uid: str
    name: str


CONFIG_PATH = 'databutton.json'


def get_databutton_config() -> ProjectConfig:
    with open(CONFIG_PATH, 'r') as f:
        config = json.load(f)
        return ProjectConfig(name=config['name'], uid=config['uid'])


def get_databutton_login_info() -> Union[LoginData, None]:
    auth_path = get_databutton_login_path()
    uids = [f for f in os.listdir(auth_path) if f.endswith('.json')]
    if len(uids) > 0:
        # Just take a random one for now
        with open(os.path.join(auth_path, uids[0])) as f:
            config = json.load(f)
            return LoginData(uid=config['uid'], refreshToken=config['refreshToken'])
    return None


def get_databutton_login_path():
    return os.path.join(Path.home(), '.config', 'databutton')


def create_databutton_config(name: str) -> ProjectConfig:
    config = ProjectConfig(name=name, uid=str(uuid.uuid4()))
    with open(CONFIG_PATH, 'w') as f:
        f.write(json.dumps(config.__dict__, indent=2))
        return config


FIREBASE_CONFIG = {
    'apiKey': 'AIzaSyAdgR9BGfQrV2fzndXZLZYgiRtpydlq8ug',
    'authDomain': 'databutton.firebaseapp.com',
    'projectId': 'databutton',
    'storageBucket': 'databutton.appspot.com',
    'databaseURL': ''
}
storage_bucket = f"https://firebasestorage.googleapis.com/v0/b/{FIREBASE_CONFIG['storageBucket']}"


def upload_archive(config: ProjectConfig, deployment_id: str):
    upload_path = f"projects/{config.uid}/{deployment_id}"
    url = f"{storage_bucket}/o?name={upload_path}"
    token = get_auth_token()
    headers = {"Authorization": f"Firebase {token}",
               "Content-type": "application/tar+gzip"}
    with open(deployment_id, 'rb') as f:
        response = requests.post(url, headers=headers, data=f)
        if not response.ok:
            raise Exception("Could not upload archive")
        metadata = {
            'deployment_id': deployment_id
        }
        meta_url = f"{storage_bucket}/o/{quote(upload_path, safe='')}"
        meta_res = requests.patch(meta_url, headers={'Authorization': headers['Authorization']}, json={'metadata': metadata})
        if not meta_res.ok:
            raise Exception("Could not update metadata")


def get_auth_token() -> str:
    login_info = get_databutton_login_info()
    res = requests.post(f"https://securetoken.googleapis.com/v1/token?key={FIREBASE_CONFIG['apiKey']}", {
        'grant_type': 'refresh_token', 'refresh_token': login_info.refreshToken})
    if not res.ok:
        raise Exception('Could not authenticate')
    json = res.json()
    return json['id_token']
